@file:OptIn(ExperimentalPagingApi::class)

package dev.adriankuta.pixabay.data.repository

import androidx.paging.ExperimentalPagingApi
import androidx.paging.Pager
import androidx.paging.PagingConfig
import androidx.paging.PagingData
import androidx.paging.map
import dev.adriankuta.pixabay.data.model.PixabayImage
import dev.adriankuta.pixabay.data.network.PixabayService
import dev.adriankuta.pixabay.data.paging.PixabayPagingSource
import dev.adriankuta.pixabay.data.paging.PixabayRemoteMediator
import dev.adriankuta.pixabay.data.room.AppDatabase
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.map
import timber.log.Timber
import javax.inject.Inject
import javax.inject.Named

internal class PixabayImageRepository @Inject constructor(
    private val database: AppDatabase,
    private val pixabayService: PixabayService
) : ImageRepository {

    override suspend fun searchImageById(id: Int): PixabayImage {
        return runCatching {
            pixabayService.searchImageById(id.toString())
        }.mapCatching { response ->
            PixabayImage(response.hits.first())
        }.onFailure { e ->
            Timber.e(e)
        }.getOrThrow()
    }

    override suspend fun searchImages(query: String, page: Int, pageSize: Int): List<PixabayImage> {
        return runCatching {
            pixabayService.searchImages(query, page, pageSize)
        }.mapCatching { response ->
            response.hits.map { PixabayImage(it) }
        }.onFailure { e ->
            Timber.e(e)
        }.getOrThrow()
    }

    override fun getSearchResultStream(query: String): Flow<PagingData<PixabayImage>> {
        return if (USE_CACHE_PAGER)
            getOnlinePagerWithCache(query)
        else
            getOnlinePager(query)
    }

    private fun getOnlinePager(query: String) = Pager(
        config = PagingConfig(
            pageSize = NETWORK_PAGE_SIZE,
            enablePlaceholders = false
        ),
        pagingSourceFactory = { PixabayPagingSource(pixabayService, query) }
    ).flow

    private fun getOnlinePagerWithCache(query: String) = Pager(
        config = PagingConfig(
            pageSize = NETWORK_PAGE_SIZE,
            enablePlaceholders = false
        ),
        remoteMediator = PixabayRemoteMediator(
            query,
            database,
            pixabayService
        ),
        pagingSourceFactory = { database.imagesDao().imagesByQuery(query) }
    ).flow
        .map { pagingSource ->
            pagingSource.map { PixabayImage(it) }
        }

    companion object {
        const val NETWORK_PAGE_SIZE = 30
        const val USE_CACHE_PAGER = false
    }
}